entity test is
    generic(
        max: integer
    );
    -- abstract begin
    -- constraint max>1;
    -- abstract end
    port(
        reset: in std_logic;
        clk: in std_logic;
        wr: in std_logic;
        rd: in std_logic;
        full: out std_logic;    -- ==1 if addr==max-1
        empty: out std_logic    -- ==1 if addr==0
    );
end entity test;

architecture behavioral of test is
    signal addr: std_logic_vector(32 downto 0);
    signal read_allow: std_logic;
    signal write_allow: std_logic;
    -- bad state specification
    signal bad: std_logic;
    signal p_addr: std_logic_vector(32 downto 0);
begin

read_allow <= (not empty and rd and not wr);
write_allow <= (not full and wr and not rd);

-- addr
process(reset, clk)
begin
    if (reset = '1') then
        addr <= (others => '0');
    elsif (clk'event and clk = '1') then
        if (read_allow = '1') then
            addr <= addr-1;
        elsif (write_allow = '1') then
            addr <= addr+1;
        end if;
    end if;
end process;

-- empty
process(reset, clk)
begin
    if (reset = '1') then
        empty <= '1';
    elsif (clk'event and clk = '1') then
        empty <= addr = 0;
    end if;
end process;

-- full
process(reset, clk)
begin
    if (reset = '1') then
        full <= '0';
    elsif (clk'event and clk = '1') then
        full <= addr = max-1;
    end if;
end process;

--------------------------------------------------------------------------
-- bad
-- Full and Empty is set appropriatelly.

-- next(p_addr) = addr
process(reset, clk)
begin
    if (reset = '1') then
        p_addr <= (others => '0');
    elsif (clk'event and clk = '1') then
        p_addr <= addr;
    end if;
end process;

-- Due to ARMC output plugin not understanding non-CNF of expression...
-- bad <= not((reset and empty and (addr=0)) or
--            (not reset and empty and (addr=0)) or
--            (not reset and full and (addr=(max-1))) or
--            (not reset and not empty and (addr>0)) or
--            (not reset and not full and (addr<(max-1))));
-- We must rewrite it:
bad <= not(reset and empty and (p_addr=0)) and
       not(not reset and empty and (p_addr=0)) and
       not(not reset and full and (p_addr=(max-1))) and
       not(not reset and not empty and (p_addr>0)) and
       not(not reset and not full and (p_addr<(max-1)));

end architecture behavioral;
