#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:ts=4:expandtab:sw=4:tw=0
###############################################################################
# output.py - output module for ARMC http://www.mpi-inf.mpg.de/~rybal/armc/   #
#             from Andrey Rybalchenko http://www.mpi-inf.mpg.de/~rybal/       #
# $Id: output_armc.py 381 2007-11-19 17:53:20Z smrcka $
#                                                                             #
# Author: Ales Smrcka <smrcka@fit.vutbr.cz>                                   #
#                                                                             #
# This program is free software; you can redistribute it and/or modify        #
#   it under the terms of the GNU General Public License as published by      #
#   the Free Software Foundation; either version 2 of the License, or         #
#   (at your option) any later version.                                       #
#   See http://www.gnu.org/licenses/gpl.html for more details.                #
#                                                                             #
###############################################################################
from simplify import strexpr

# options
TRANSFORM_INEQ=1          # a!=b --> a<b or a>b
TRANSFORM_STRICT_INEQ=1   # a>b --> a>=b+1

class ConvError(Exception):
    pass

def varname(name, case='P'):
    return '%s_%s' % (case, name)

varmap={}
nvarmap={}

# transition counter
trcounter = 1
def tr(f, t, pre_predicates, post_predicates, vars, nvars):
    global trcounter
    s ='r(p(pc(%s), data(%s)),\n' % (f,vars)
    s+='  p(pc(%s), data(%s)),\n' % (t,nvars)
    s+='  [%s],\n' % pre_predicates
    s+='  [%s],\n' % post_predicates
    s+='  %d).\n\n' % trcounter
    trcounter += 1
    return s

def transition2str(t, guards, actions, vars, nvars):
    transition = ''
    for g in guards:
        if type(g) == list and g[0] == '!=':
            g[0] = '<'
            transition += transition2str(t, guards, actions, vars, nvars)
            g[0] = '>'
            transition += transition2str(t, guards, actions, vars, nvars)
            return transition
    guards = [expr2str(g, varmap) for g in guards]
#TESTING
#    predicates = nice_list(guards+actions, ',', '   ', 1)
#    return tr(t.s1.id, t.s2.id, predicates, vars, nvars)
    pre_predicates = nice_list(guards, ',', '   ', 1)
    post_predicates = nice_list(actions, ',', '   ', 1)
    return tr(t.s1.id, t.s2.id, pre_predicates, post_predicates, vars, nvars)


def nice_list(l, delimiter, indent, items_per_line):
    tmp=l[:]
    c=0
    for i in range(len(tmp)):
        if i and i%items_per_line==0:
            tmp[i]='\n'+indent+tmp[i]
    return delimiter.join(tmp)

def transformate_automaton(ca):
    from time import strftime
    from sys import argv
    model=header % ('trans.py', ca.name, ','.join(ca.sources), ca.time, \
        strftime('%c'), ' '.join(argv))
    model+='% variables\n'
    variables=list(ca.counters|ca.parameters)
    varnames=[]
    # create variable map (Variable->armc identificator)
    for v in variables:
        varmap[v] = varname(v, 'P')
        nvarmap[v] = varname(v, 'N')
        varnames.append(v)
    vars=nice_list([varname(name, 'P') for name in variables], ',', ' '*12, 3)
    nvars=nice_list([varname(name, 'N') for name in variables], ',', ' '*12, 3)
    tmp=nice_list(["(%s,'%s')" % (varmap[v], v.name) for v in variables], \
        ', ', ' '*11, 1)
    model+='var2names(p(_, data(%s)),\n          [%s]).\n\n' % (vars, tmp)

    model+='%' + '-'*75 + '\n'
    model+='preds(p(_, data(%s)), []).\n' % vars
    model+='%' + '-'*75 + '\n\n'

    model+='cube_size(1).\n\n'
    model+='implicit_updates.\n\n'

    model+='start(pc(%s)).\n' % ca.init_state.id
    model+='error(pc(%s)).\n\n' % ca.bad_state.id
    
    # init state first
    states = list(ca.states)
    del states[states.index(ca.init_state)]
    states = [ca.init_state]+states

    # generation of transitions
    for s in states:
        otransitions=[t for t in ca.transitions if t.s1==s]
        for t in otransitions:
            model+='%' + '-'*75 + '\n'
            model+='%% Transition %s->%s\n' % (s.id, t.s2.id)
            model+='%% state %s: %s\n' % (s.id, s)
            model+='%% state %s: %s\n' % (t.s2.id, t.s2)
            model+='%' + '-'*75 + '\n'
            guards = guard2list(t.guard)
            if t == ca.init_transition:
                actions = armc_actions(t.actions, ca.parameters) + \
                          armc_constraints( \
                    [p.comment for p in ca.parameters if p.comment])
            else:
                actions = armc_actions(t.actions, [])
            model+=transition2str(t, guards, actions, vars, nvars)

    return model 

# transitional state counter
ctmp = 0
def get_trstate():
    global ctmp
    ctmp += 1
    return 'trstate%d' % ctmp

opmap={
    '-':' - ',
    '+':' + ',
    '==':' = ',
    '!=':' =\\= ',
    '<':' < ',
    '>':' > ',
    '<=':' =< ',
    '>=':' >= ',
    '*':' * ',
}

def expr2str(expr, vmap):
    if type(expr)==int:
        return str(expr)
    elif type(expr)==list:
        if expr[0] in ('and', 'or', 'not'):
            raise ConvError('ARMC:unsupported and,or,not in expression: %s' \
                % strexpr(expr))
        else:
            if len(expr)>2:
                if TRANSFORM_STRICT_INEQ:
                    # a>b --> a>=b+1, a<b --> a<=b-1
                    if expr[0]=='>':
                        assert len(expr)==3
                        return '(%s >= %s+1)' % \
                            (expr2str(expr[1], vmap), expr2str(expr[2], vmap))
                    elif expr[0]=='<':
                        assert len(expr)==3
                        return '(%s =< %s-1)' % \
                            (expr2str(expr[1], vmap), expr2str(expr[2], vmap))
                    else:
                        return '(%s)' % opmap[expr[0]].join( \
                            [expr2str(e, vmap) for e in expr[1:]])
                else:
                    return '(%s)' % opmap[expr[0]].join( \
                        [expr2str(e, vmap) for e in expr[1:]])
            else: # unary operator
                return '(%s%s)' % (opmap[expr[0]], expr2str(expr[1], vmap))
    else: # VariableUse
        return vmap[expr.v]

def guard2list(guard):
    if type(guard) == list:
        if guard[0] == 'and':
            return guard[1:]
        else:
            return [guard]
    elif guard != 1:
        return [guard]
    else:
        return []

def armc_guard(guard):
    if guard==1:
        return []
    else:
        expr = armc_expr(guard, True)
        if type(expr) == list:
            return expr
        else:
            return [expr]

def armc_actions(actions, parameters):
    # set the new value for every counter
    new_values = ['%s = %s' % (nvarmap[a.counter], expr2str(a.value, varmap)) \
                    for a in actions]
    counters_preserving_value = []
    new_counters = [a.counter for a in actions] + list(parameters)
    for counter in nvarmap.keys():
        if counter not in new_counters:
            new_values.append('%s = %s' % (nvarmap[counter], varmap[counter]))
    return new_values
#    return ['%s = %s' % (nvarmap[a.counter], expr2str(a.value, varmap)) \
#                for a in actions]

def armc_constraints(constraints):
    return [expr2str(c, nvarmap) for c in constraints]

header="""%% This model was automatically generated by %s
%% For more information contact: Ales Smrcka <smrcka@fit.vutbr.cz>
%% model name: %s
%% source files: %s
%% transformation time: %s s
%% created: %s
%% cmd line: %s

:- multifile r/5, implicit_updates/0, var2names/2, preds/2, cube_size/1, start/1, refinement/1, error/1.

refinement(inter).

"""
