#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:ts=4:expandtab:sw=4:tw=0
###############################################################################
# output.py - output modul for .dot format                                    #
#                                                                             #
# $Id: output_fast.py 170 2007-05-03 15:03:53Z smrcka $
#                                                                             #
# Author: Ales Smrcka <smrcka@fit.vutbr.cz>                                   #
#                                                                             #
# This program is free software; you can redistribute it and/or modify        #
#   it under the terms of the GNU General Public License as published by      #
#   the Free Software Foundation; either version 2 of the License, or         #
#   (at your option) any later version.                                       #
#   See http://www.gnu.org/licenses/gpl.html for more details.                #
#                                                                             #
###############################################################################

def transformate_automaton(ca):
    from time import strftime
    from sys import argv
    model=header % ('trans.py', ca.name, ','.join(ca.sources), ca.time, \
        strftime('%c'), ' '.join(argv))
    model+='digraph %s {\n' % ca.name
    model+='    rankdir=LR;\n'
    model+='    size="20,20";\n'
    model+='    node [style=filled fillcolor=greenyellow width=0.7 height=0.7];\n'
    model+='    %s [style=filled fillcolor=cyan];\n' % ca.init_state.id
    if ca.bad_state:
        model+='    %s [style=filled fillcolor=red];\n' % ca.bad_state.id
    # simple transitions
    t_counter = 1
    labels = []
    for t in ca.transitions:
        if t.guard==1:
            guard=''
        else:
            guard='%s: ' % infix(t.guard)
        sa = ["%s:=%s" % (a.counter.name, infix(a.value)) for a in t.actions]
        if t == ca.init_transition:
            sa += [infix(p.comment) for p in ca.parameters if p.comment]
        if sa:
            actions=', '.join(sa)
        else:
            actions=''
        label = guard + actions
        labels.append(label)
        model+='    %s -> %s [label="%d"];\n' % (t.s1.id, t.s2.id, t_counter)
        t_counter += 1
    model+='}\n'
    model += '/* transition labels:\n'
    t_counter = 1
    for label in labels:
        if not label: label = 'true'
        model += '%d = %s\n' % (t_counter, label)
        t_counter += 1
    model += '*/\n'
    return model 

opmap={
    'not':'!',
    'and':' && ',
    'or':' || ',
    '-':' - ',
    '+':' + ',
    '==':' = ',
    '<':' < ',
    '>':' > ',
    '<=':' <= ',
    '>=':' >= ',
}

def infix(expr):
    if type(expr)==int:
        return str(expr)
    elif type(expr)==list:
        if expr[0]=='!=':
            return infix(['not',['==']+expr[1:]])
        else:
            if len(expr)>2:
                return '(%s)' % (opmap[expr[0]].join( \
                    [infix(o) for o in expr[1:]]))
            else:
                return '(%s%s)' % (opmap[expr[0]], infix(expr[1]))
    else: # VariableUse
        return str(expr.v)

header="""/* This model was automatically generated by %s
For more information contact: Ales Smrcka <smrcka@fit.vutbr.cz>
model name: %s
source files: %s
transformation time: %s s
created: %s
cmd line: %s
*/
"""
