#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:ts=4:expandtab:sw=4:tw=0
###############################################################################
# output.py - output modul for FAST                                           #
#                                                                             #
# $Id: output_fast.py 328 2007-09-10 09:22:26Z smrcka $
#                                                                             #
# Author: Ales Smrcka <smrcka@fit.vutbr.cz>                                   #
#                                                                             #
# This program is free software; you can redistribute it and/or modify        #
#   it under the terms of the GNU General Public License as published by      #
#   the Free Software Foundation; either version 2 of the License, or         #
#   (at your option) any later version.                                       #
#   See http://www.gnu.org/licenses/gpl.html for more details.                #
#                                                                             #
###############################################################################

def transformate_automaton(ca):
    model=header % ('trans.py', ','.join(ca.sources), ca.time, ca.name)
    model+='model %s {\n' % ca.name
    model+='var %s;\n' % ','.join(['null']+ \
        [c.name for c in ca.counters|ca.parameters])
    model+='states %s;\n\n' % ','.join([s.id for s in ca.states])
    ts=[]
    i=0
    for t in ca.transitions:
        model+='transition t%d := {\n' % i
        model+='    from := %s;\n' % t.s1.id
        model+='    to := %s;\n' % t.s2.id
        if t.guard==1:
            model+='    guard := true;\n'
        else:
            model+='    guard := %s;\n' % infix_fast(t.guard)
        sa=["%s'=%s" % (a.counter.name,infix_fast(a.value)) for a in t.actions]
        sa=', '.join(sa)
        if sa:
            model+='    action := %s;\n' % sa
        else:
            model+="    action := null'=0;\n"
        model+='};\n\n'
        ts.append('t%d' % i)
        i+=1
    model+='}\n\nstrategy %s {\n\n' % ca.name
    model+='/* state summary */\n'
    states=list(ca.states)
    states.sort()
    for s in states:
        model+='/* state %5s: %s */\n' % (s.id,s)
    model+='\n'
    model+='/* state variable summary */\n'
    sv=list(ca.state_vars)
    sv.sort()
    for v in sv:
        model+='\n/* %s: %s */\n' % (v.name, ','.join( \
            [s.id for s in ca.states if s.v[v.name]]))
    model+='\n'
    if ca.init_state:
        init=ca.init_state
    else:
        init='s0 && %s' % ' && '.join(['%s=0' % c.name for c in ca.counters])
    model+='Region init := {state=%s};\n\n' % init
    model+='Transitions t := {%s};\n\n' % ','.join(ts)
    model+='Region reach := post*(init, t);\n\n'
    model+='Region good := {};\n\n'
    model+='boolean result := isEmpty(reach && good);\n'
    model+='if result then\n    print ("Property does not hold.");\n'
    model+='else\n    print ("Property hold.");\nendif\n'
    model+='}\n'
    return model 

opmap={
    'not':'!',
    'and':'&&',
    'or':'||',
    '-':'-',
    '+':'+',
    '==':'=',
    '<':'<',
    '>':'>',
    '<=':'<=',
    '>=':'>=',
}

def infix_fast(expr):
    if type(expr)==int:
        return str(expr)
    elif type(expr)==list:
        if expr[0]=='!=':
            return infix_fast(['not',['==']+expr[1:]])
        else:
            if len(expr)>2:
                return '(%s)' % (opmap[expr[0]].join( \
                    [infix_fast(o) for o in expr[1:]]))
            else:
                return '(%s%s)' % (opmap[expr[0]], infix_fast(expr[1]))
    else: # VariableUse
        return str(expr.v)
    

header="""/*
***************************************************************************
This model was automatically generated by %s
For more information contact: Ales Smrcka <smrcka@fit.vutbr.cz>
source files: %s
transformation time: %s
model name: %s
**************************************************************************/
"""
